@@ -19,6 +19,7 @@ import pyarrow as pa
 import time
 import re
 import io
+import logging
 from enum import Enum, unique
 from typing import List, Union
 from requests_toolbelt import multipart
@@ -34,6 +35,9 @@ PATH_TRANSACTIONS = "/transactions"
 PATH_USER = "/users"
 PATH_OAUTH_CLIENT = "/oauth-clients"
 
+# logger
+logger = logging.getLogger(__package__)
+
 
 # Engine sizes
 @unique
@@ -71,6 +75,11 @@ class Permission(str, Enum):
     DELETE_COMPUTE = "delete:compute"
     LIST_COMPUTES = "list:compute"
     READ_COMPUTE = "read:compute"
+    # engines (duplicate to computes)
+    CREATE_ENGINE = "create:engine"
+    DELETE_ENGINE = "delete:engine"
+    LIST_ENGINES = "list:engine"
+    READ_ENGINE = "read:engine"
     # databases
     LIST_DATABASES = "list:database"
     UPDATE_DATABASE = "update:database"
@@ -325,27 +334,73 @@ def _parse_arrow_results(files: List[TransactionAsyncFile]):
             results.append({"relationId": file.name, "table": table})
     return results
 
+# polling with specified overhead
+# delay is the overhead % of the time the transaction has been running so far
 
-def create_engine(ctx: Context, engine: str, size: EngineSize = EngineSize.XS):
+
+def poll_with_specified_overhead(
+    f,
+    overhead_rate: float,
+    start_time: int = time.time(),
+    timeout: int = None,
+    max_tries: int = None,
+    max_delay: int = 120,
+):
+    tries = 0
+    max_time = time.time() + timeout if timeout else None
+
+    while True:
+        if f():
+            break
+
+        if max_tries is not None and tries >= max_tries:
+            raise Exception(f'max tries {max_tries} exhausted')
+
+        if max_time is not None and time.time() >= max_time:
+            raise Exception(f'timed out after {timeout} seconds')
+
+        tries += 1
+        duration = min((time.time() - start_time) * overhead_rate, max_delay)
+        if tries == 1:
+            time.sleep(0.5)
+        else:
+            time.sleep(duration)
+
+
+def is_engine_term_state(state: str) -> bool:
+    return state == "PROVISIONED" or ("FAILED" in state)
+
+
+def create_engine(ctx: Context, engine: str, size: EngineSize = EngineSize.XS, **kwargs):
     data = {"region": ctx.region, "name": engine, "size": size.value}
     url = _mkurl(ctx, PATH_ENGINE)
-    rsp = rest.put(ctx, url, data)
+    rsp = rest.put(ctx, url, data, **kwargs)
     return json.loads(rsp.read())
 
 
-def create_user(ctx: Context, email: str, roles: List[Role] = None):
+def create_engine_wait(ctx: Context, engine: str, size: EngineSize = EngineSize.XS, **kwargs):
+    create_engine(ctx, engine, size, **kwargs)
+    poll_with_specified_overhead(
+        lambda: is_engine_term_state(get_engine(ctx, engine)["state"]),
+        overhead_rate=0.2,
+        timeout=30 * 60,
+    )
+    return get_engine(ctx, engine)
+
+
+def create_user(ctx: Context, email: str, roles: List[Role] = None, **kwargs):
     rs = roles or []
     data = {"email": email, "roles": [r.value for r in rs]}
     url = _mkurl(ctx, PATH_USER)
-    rsp = rest.post(ctx, url, data)
+    rsp = rest.post(ctx, url, data, **kwargs)
     return json.loads(rsp.read())
 
 
-def create_oauth_client(ctx: Context, name: str, permissions: List[Permission] = None):
+def create_oauth_client(ctx: Context, name: str, permissions: List[Permission] = None, **kwargs):
     ps = permissions or []
     data = {"name": name, "permissions": ps}
     url = _mkurl(ctx, PATH_OAUTH_CLIENT)
-    rsp = rest.post(ctx, url, data)
+    rsp = rest.post(ctx, url, data, **kwargs)
     return json.loads(rsp.read())["client"]
 
 
@@ -358,60 +413,60 @@ def _create_mode(source_database: str, overwrite: bool) -> Mode:
     return result
 
 
-def delete_database(ctx: Context, database: str) -> dict:
+def delete_database(ctx: Context, database: str, **kwargs) -> dict:
     data = {"name": database}
     url = _mkurl(ctx, PATH_DATABASE)
-    rsp = rest.delete(ctx, url, data)
+    rsp = rest.delete(ctx, url, data, **kwargs)
     return json.loads(rsp.read())
 
 
-def delete_engine(ctx: Context, engine: str) -> dict:
+def delete_engine(ctx: Context, engine: str, **kwargs) -> dict:
     data = {"name": engine}
     url = _mkurl(ctx, PATH_ENGINE)
-    rsp = rest.delete(ctx, url, data)
+    rsp = rest.delete(ctx, url, data, **kwargs)
     return json.loads(rsp.read())
 
 
-def delete_user(ctx: Context, id: str) -> dict:
+def delete_user(ctx: Context, id: str, **kwargs) -> dict:
     url = _mkurl(ctx, f"{PATH_USER}/{id}")
-    rsp = rest.delete(ctx, url, None)
+    rsp = rest.delete(ctx, url, None, **kwargs)
     return json.loads(rsp.read())
 
 
-def disable_user(ctx: Context, userid: str) -> dict:
-    return update_user(ctx, userid, status="INACTIVE")
+def disable_user(ctx: Context, userid: str, **kwargs) -> dict:
+    return update_user(ctx, userid, status="INACTIVE", **kwargs)
 
 
-def delete_oauth_client(ctx: Context, id: str) -> dict:
+def delete_oauth_client(ctx: Context, id: str, **kwargs) -> dict:
     url = _mkurl(ctx, f"{PATH_OAUTH_CLIENT}/{id}")
-    rsp = rest.delete(ctx, url, None)
+    rsp = rest.delete(ctx, url, None, **kwargs)
     return json.loads(rsp.read())
 
 
-def enable_user(ctx: Context, userid: str) -> dict:
-    return update_user(ctx, userid, status="ACTIVE")
+def enable_user(ctx: Context, userid: str, **kwargs) -> dict:
+    return update_user(ctx, userid, status="ACTIVE", **kwargs)
 
 
-def get_engine(ctx: Context, engine: str) -> dict:
-    return _get_resource(ctx, PATH_ENGINE, name=engine, deleted_on="", key="computes")
+def get_engine(ctx: Context, engine: str, **kwargs) -> dict:
+    return _get_resource(ctx, PATH_ENGINE, name=engine, deleted_on="", key="computes", **kwargs)
 
 
-def get_database(ctx: Context, database: str) -> dict:
-    return _get_resource(ctx, PATH_DATABASE, name=database, key="databases")
+def get_database(ctx: Context, database: str, **kwargs) -> dict:
+    return _get_resource(ctx, PATH_DATABASE, name=database, key="databases", **kwargs)
 
 
-def get_oauth_client(ctx: Context, id: str) -> dict:
-    return _get_resource(ctx, f"{PATH_OAUTH_CLIENT}/{id}", key="client")
+def get_oauth_client(ctx: Context, id: str, **kwargs) -> dict:
+    return _get_resource(ctx, f"{PATH_OAUTH_CLIENT}/{id}", key="client", **kwargs)
 
 
-def get_transaction(ctx: Context, id: str) -> dict:
-    return _get_resource(ctx, f"{PATH_TRANSACTIONS}/{id}", key="transaction")
+def get_transaction(ctx: Context, id: str, **kwargs) -> dict:
+    return _get_resource(ctx, f"{PATH_TRANSACTIONS}/{id}", key="transaction", **kwargs)
 
 
-def get_transaction_metadata(ctx: Context, id: str) -> list:
+def get_transaction_metadata(ctx: Context, id: str, **kwargs) -> list:
     headers = {"Accept": "application/x-protobuf"}
     url = _mkurl(ctx, f"{PATH_TRANSACTIONS}/{id}/metadata")
-    rsp = rest.get(ctx, url, headers=headers)
+    rsp = rest.get(ctx, url, headers=headers, **kwargs)
     content_type = rsp.headers.get("content-type", "")
     if "application/x-protobuf" in content_type:
         return _parse_metadata_proto(rsp.read())
@@ -419,17 +474,17 @@ def get_transaction_metadata(ctx: Context, id: str) -> list:
     raise Exception(f"invalid content type for metadata proto: {content_type}")
 
 
-def list_transactions(ctx: Context) -> list:
-    return _get_collection(ctx, PATH_TRANSACTIONS, key="transactions")
+def list_transactions(ctx: Context, **kwargs) -> list:
+    return _get_collection(ctx, PATH_TRANSACTIONS, key="transactions", **kwargs)
 
 
-def get_transaction_problems(ctx: Context, id: str) -> list:
-    return _get_collection(ctx, f"{PATH_TRANSACTIONS}/{id}/problems")
+def get_transaction_problems(ctx: Context, id: str, **kwargs) -> list:
+    return _get_collection(ctx, f"{PATH_TRANSACTIONS}/{id}/problems", **kwargs)
 
 
-def get_transaction_results(ctx: Context, id: str) -> list:
+def get_transaction_results(ctx: Context, id: str, **kwargs) -> list:
     url = _mkurl(ctx, f"{PATH_TRANSACTIONS}/{id}/results")
-    rsp = rest.get(ctx, url)
+    rsp = rest.get(ctx, url, **kwargs)
     content_type = rsp.headers.get("content-type", "")
     if "multipart/form-data" in content_type:
         parts = _parse_multipart_form(content_type, rsp.read())
@@ -442,20 +497,20 @@ def get_transaction_results(ctx: Context, id: str) -> list:
 # deprecated, get_transaction_results should be called instead
 
 
-def get_transaction_results_and_problems(ctx: Context, id: str) -> list:
+def get_transaction_results_and_problems(ctx: Context, id: str, **kwargs) -> list:
     rsp = TransactionAsyncResponse()
-    rsp.problems = get_transaction_problems(ctx, id)
-    rsp.results = get_transaction_results(ctx, id)
+    rsp.problems = get_transaction_problems(ctx, id, **kwargs)
+    rsp.results = get_transaction_results(ctx, id, **kwargs)
     return rsp
 
 
-def cancel_transaction(ctx: Context, id: str) -> dict:
-    rsp = rest.post(ctx, _mkurl(ctx, f"{PATH_TRANSACTIONS}/{id}/cancel"), {})
+def cancel_transaction(ctx: Context, id: str, **kwargs) -> dict:
+    rsp = rest.post(ctx, _mkurl(ctx, f"{PATH_TRANSACTIONS}/{id}/cancel"), {}, **kwargs)
     return json.loads(rsp.read())
 
 
-def get_user(ctx: Context, userid: str) -> dict:
-    return _get_resource(ctx, f"{PATH_USER}/{userid}", name=userid)
+def get_user(ctx: Context, userid: str, **kwargs) -> dict:
+    return _get_resource(ctx, f"{PATH_USER}/{userid}", name=userid, **kwargs)
 
 
 def list_engines(ctx: Context, state=None) -> list:
@@ -472,22 +527,22 @@ def list_databases(ctx: Context, state=None) -> list:
     return _get_collection(ctx, PATH_DATABASE, key="databases", **kwargs)
 
 
-def list_users(ctx: Context) -> list:
-    return _get_collection(ctx, PATH_USER, key="users")
+def list_users(ctx: Context, **kwargs) -> list:
+    return _get_collection(ctx, PATH_USER, key="users", **kwargs)
 
 
-def list_oauth_clients(ctx: Context) -> list:
-    return _get_collection(ctx, PATH_OAUTH_CLIENT, key="clients")
+def list_oauth_clients(ctx: Context, **kwargs) -> list:
+    return _get_collection(ctx, PATH_OAUTH_CLIENT, key="clients", **kwargs)
 
 
-def update_user(ctx: Context, userid: str, status: str = None, roles=None):
+def update_user(ctx: Context, userid: str, status: str = None, roles=None, **kwargs):
     data = {}
     if status is not None:
         data["status"] = status
     if roles is not None:
         data["roles"] = roles
     url = _mkurl(ctx, f"{PATH_USER}/{userid}")
-    rsp = rest.patch(ctx, url, data)
+    rsp = rest.patch(ctx, url, data, **kwargs)
     return json.loads(rsp.read())
 
 
@@ -578,13 +633,14 @@ class TransactionAsync(object):
             result["engine_name"] = self.engine
         return result
 
-    def run(self, ctx: Context, command: str, inputs: dict = None) -> Union[dict, list]:
+    def run(self, ctx: Context, command: str, language: str, inputs: dict = None, **kwargs) -> Union[dict, list]:
         data = self.data
         data["query"] = command
+        data["language"] = language
         if inputs is not None:
             inputs = [_query_action_input(k, v) for k, v in inputs.items()]
             data["v1_inputs"] = inputs
-        rsp = rest.post(ctx, _mkurl(ctx, PATH_TRANSACTIONS), data)
+        rsp = rest.post(ctx, _mkurl(ctx, PATH_TRANSACTIONS), data, **kwargs)
         content_type = rsp.headers.get("content-type", None)
         content = rsp.read()
         # todo: response model should be based on status code (200 v. 201)
@@ -667,10 +723,10 @@ def _list_models(ctx: Context, database: str, engine: str) -> dict:
     return models
 
 
-def create_database(ctx: Context, database: str, source: str = None) -> dict:
+def create_database(ctx: Context, database: str, source: str = None, **kwargs) -> dict:
     data = {"name": database, "source_name": source}
     url = _mkurl(ctx, PATH_DATABASE)
-    rsp = rest.put(ctx, url, data)
+    rsp = rest.put(ctx, url, data, **kwargs)
     return json.loads(rsp.read())
 
 
@@ -833,24 +889,32 @@ def exec(
     command: str,
     inputs: dict = None,
     readonly: bool = True,
+    **kwargs
 ) -> TransactionAsyncResponse:
+    logger.info('exec: database %s engine %s readonly %s' % (database, engine, readonly))
+    start_time = time.time()
     txn = exec_async(ctx, database, engine, command, inputs=inputs, readonly=readonly)
     # in case of if short-path, return results directly, no need to poll for
     # state
     if not (txn.results is None):
+        logger.info('transaction id: %s' % txn.transaction["id"])
         return txn
 
     rsp = TransactionAsyncResponse()
-    txn = get_transaction(ctx, txn.transaction["id"])
-    while True:
-        time.sleep(1)
-        txn = get_transaction(ctx, txn["id"])
-        if is_txn_term_state(txn["state"]):
-            rsp.transaction = txn
-            rsp.metadata = get_transaction_metadata(ctx, txn["id"])
-            rsp.problems = get_transaction_problems(ctx, txn["id"])
-            rsp.results = get_transaction_results(ctx, txn["id"])
-            break
+    txn = get_transaction(ctx, txn.transaction["id"], **kwargs)
+
+    poll_with_specified_overhead(
+        lambda: is_txn_term_state(get_transaction(ctx, txn["id"], **kwargs)["state"]),
+        overhead_rate=0.2,
+        start_time=start_time,
+    )
+
+    logger.info('transaction id: %s' % txn["id"])
+
+    rsp.transaction = get_transaction(ctx, txn["id"], **kwargs)
+    rsp.metadata = get_transaction_metadata(ctx, txn["id"], **kwargs)
+    rsp.problems = get_transaction_problems(ctx, txn["id"], **kwargs)
+    rsp.results = get_transaction_results(ctx, txn["id"], **kwargs)
 
     return rsp
 
@@ -862,9 +926,11 @@ def exec_async(
     command: str,
     readonly: bool = True,
     inputs: dict = None,
+    language: str = "",
+    **kwargs,
 ) -> TransactionAsyncResponse:
     tx = TransactionAsync(database, engine, readonly=readonly)
-    rsp = tx.run(ctx, command, inputs=inputs)
+    rsp = tx.run(ctx, command, language=language, inputs=inputs, **kwargs)
 
     if isinstance(rsp, dict):
         return TransactionAsyncResponse(rsp, None, None, None)
