@@ -18,7 +18,6 @@ import time
 
 __all__ = [
     "Credentials",
-    "AccessKeyCredentials",
     "AccessToken",
     "ClientCredentials",
 ]
@@ -31,19 +30,13 @@ class Credentials(ABC):
     pass
 
 
-# Represents access key credentials.
-class AccessKeyCredentials(Credentials):
-    def __init__(self, akey: str, pkey: str):
-        self.akey = akey  # access_key
-        self.pkey = pkey  # private_key
-
-
 # Represents an OAuth access token.
 class AccessToken:
-    def __init__(self, access_token: str, expires_in: int):
-        self.token = access_token
+    def __init__(self, access_token: str, scope: str, expires_in: int, created_on: float = time.time()):
+        self.access_token = access_token
+        self.scope = scope
         self.expires_in = expires_in
-        self.created_on = round(time.time())
+        self.created_on = created_on
 
     def is_expired(self):
         return (
